/*
 * Stereo Cross Delay audio effect based on DISTRHO Plugin Framework (DPF)
 *
 * SPDX-License-Identifier: MIT
 *
 * Copyright (C) 2020 Christopher Arndt <info@chrisarndt.de>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

#ifndef PLUGIN_STEREOCROSSDELAY_H
#define PLUGIN_STEREOCROSSDELAY_H

#include "DistrhoPlugin.hpp"
#include "StereoCrossDelay.hpp"

START_NAMESPACE_DISTRHO

#ifndef MIN
#define MIN(a,b) ( (a) < (b) ? (a) : (b) )
#endif

#ifndef MAX
#define MAX(a,b) ( (a) > (b) ? (a) : (b) )
#endif

#ifndef CLAMP
#define CLAMP(v, min, max) (MIN((max), MAX((min), (v))))
#endif

// -----------------------------------------------------------------------

class PluginStereoCrossDelay : public Plugin {
public:
    enum Parameters {
        p_delay_l,
        p_delay_r,
        p_feedback,
        p_lp_cutoff,
        p_hp_cutoff,
        p_crossmix,
        p_dry,
        p_wet,
        p_bypass_process,
        paramCount
    };

    PluginStereoCrossDelay();

    ~PluginStereoCrossDelay();

protected:
    // -------------------------------------------------------------------
    // Information

    const char* getLabel() const noexcept override {
        return "Stereo Cross-Delay";
    }

    const char* getDescription() const override {
        return "A stereo delay with feedback and cross-mixing";
    }

    const char* getMaker() const noexcept override {
        return "chrisarndt.de";
    }

    const char* getHomePage() const override {
        return DISTRHO_PLUGIN_URI;
    }

    const char* getLicense() const noexcept override {
        return "https://spdx.org/licenses/MIT";
    }

    uint32_t getVersion() const noexcept override {
        return d_version(0, 2, 0);
    }

    // Go to:
    //
    // http://service.steinberg.de/databases/plugin.nsf/plugIn
    //
    // Get a proper plugin UID and fill it in here!
    int64_t getUniqueId() const noexcept override {
        return d_cconst('S', 't', 'X', 'D');
    }

    // -------------------------------------------------------------------
    // Init

    void initParameter(uint32_t index, Parameter& parameter) override;
    void initProgramName(uint32_t index, String& programName) override;

    // -------------------------------------------------------------------
    // Internal data

    float getParameterValue(uint32_t index) const override;
    void setParameterValue(uint32_t index, float value) override;
    void loadProgram(uint32_t index) override;

    // -------------------------------------------------------------------
    // Optional

    // Optional callback to inform the plugin about a sample rate change.
    void sampleRateChanged(double newSampleRate) override;

    // -------------------------------------------------------------------
    // Process

    void activate() override;
    void run(const float**, float** outputs, uint32_t frames) override;

    // -------------------------------------------------------------------

private:
    StereoCrossDelay* dsp;
    // sample rate
    double sample_rate;
    bool sr_changed;
    // enable/bypass ramping
    bool needs_ramp_down;
    bool needs_ramp_up;
    float ramp_down;
    float ramp_up;
    float ramp_up_step;
    float ramp_down_step;
    bool state_bypass;
    bool old_bypass;
    float param_bypass;

    DISTRHO_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR(PluginStereoCrossDelay)
};

struct Preset {
    const char* name;
    float params[PluginStereoCrossDelay::paramCount];
};

const Preset factoryPresets[] = {
    //,{
    //    "Preset name",  // preset name
    //    {0.0f, ...}     // array of PluginStereoCrossDelay::paramCount float param values
    //}
    {
        "Default",
        {
            250.0,    // p_delay_l
            0.75,     // p_delay_r
            25.0,     // p_feedback
            12000.0,  // p_lp_cutoff
            60.0,     // p_hp_cutoff
            25.0,     // p_crossmix
            0.0,      // p_dry
            -6.0,     // p_wet
            0.0       // p_bypass_process
        }
    },
    {"Widener", {80.0, 0.95, 0.0, 3000.0, 80.0, 25.0, 0.0, -7.5, 0.0}},
    {"Lead L/R", {333.34, 1.5, 40.0, 1000.0, 100.0, 0.0, 0.0, -12.0, 0.0}}
};

const uint presetCount = sizeof(factoryPresets) / sizeof(Preset);

// -----------------------------------------------------------------------

END_NAMESPACE_DISTRHO

#endif  // #ifndef PLUGIN_STEREOCROSSDELAY_H
